#!/usr/bin/env python
# -*- coding: utf-8; tab-width: 4; indent-tabs-mode: nil -*-
'''check synchronization of yum repositories

License:
--------
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
'''

__author__  = 'Tony Nelson'
__email__   = 'tonynelson@georgeanelson.com'
__url__     = 'http://georgeanelson.com/reposinsync.htm'
__copyright__ = 'Copyright 2006 by George A. Nelson.  All rights reserved.'
__license__ = 'GPL'
__date__    = '11 June 2006'
__version__ = '0.1'

#   The timestamps file:    in order saved      ~/.$us.data
#   *               repo_name       base_url        mirror_url...
#   fetch_time      time_delta      primary_time    primary_sum     url
#
#   "*" marks mirror list entry
#       keyed on base_url as repo_name might not be unique (?)
#       replaces the previous list, to allow for adding or deleting mirrors
#       read into:
#           mirror2base: {url: base_url, ...}
#               base repo: url maps to same url
#           base2repo: {base_url: [repo_name, {mirrors}], ...}
#           both dicts are shared by all repos
#
#   other lines are repo updates, read into:
#       transactions: {url: [[fetch_time, time_delta, primary_time, primary_sum, etc.], ...]}
#       times as timestamps or in seconds
#   scans always read base repomd
#       only read mirror repomd if not up to date
#           (i.e., current primary_time is earlier than base_url current primary_time)
#           current primary_time is the one at the end of the list
#   if new (base or mirror) primary_time later than current
#       append new list and file entries 
#   repo membership is known from mirrorlist, which must be read or cached (?)
#       a repo name might be re-used, while the URLs should be unique
#
#   time_delta can be calculated from a mirror's primary time and the base_url current primary_time
#   
#   analysis:
#       "groups" with matching times
#           at each check?  umm, what?
#           current groups already done
#       median elapsed time to update a mirror
#           make list of elapsed times
#               estimate time of update as half-way between fetch_time and base_url previous fetch_time
#                   which is the last time we checked that the mirror wasn't up-to-date.
#           sort, take middle elt
#       median 
#       frozen mirrors

import os
import sys
import re
#import urllib
from xml.dom.minidom import parseString
from glob import glob
from fnmatch import fnmatch
import tempfile
import datetime
import time
import urllib2
import socket
import threading
import signal
import urlparse


us = os.path.splitext(os.path.basename(sys.argv[0]))[0]
usdir = os.path.abspath(os.path.dirname(__file__))

def _(str): return str
#import gettext
#transdir = os.path.join(usdir, 'locale')
#gettext.install(us, transdir)

REPOFILE = '/etc/yum.repos.d/'
RELEASEVER = '5'
BASEARCH = 'i386'
VERBOSITY = 1
TIMEOUT = 20

releasever = RELEASEVER
basearch = BASEARCH
verbosity = VERBOSITY
timeout = TIMEOUT

CACHEMAPPREFIX = 'cache_map.'
CACHEDATAPREFIX = 'data.'
CACHEMIRRORPREFIX = 'mirror.'
cache_map = None
cache_dir = None
cache_map_file = None

# timestamps stuff
TIMESTAMPFILE = '~/.$us.data'
timestamp_file = None
mirror2base = {}
base2repo = {}
transactions = {}


def timestamps_load(fn):
    '''Load timestamp data and mirror -> repo map from timestamp file.
    '''
    global timestamp_file
    if not os.path.exists(fn):
        timestamp_file = open(fn, 'w')
        return
    
    timestamp_file = open(fn, 'r+')
    for l in timestamp_file:
        l = l[:-1].split('\t')
        if l[0] == '*':
            repo_name = l[1]
            base_url = l[2]
            mirrors = l[3:]
            mirrors_dict = {}.fromkeys(mirrors, base_url)
            base2repo[base_url] = [repo_name, mirrors_dict]
            mirror2base.update(mirrors_dict)
        else:
            l[0] = int(l[0])
            l[1] = int(l[1])
            l[2] = int(l[2])
            transactions.setdefault(l[-1], []).append(l[:-1])

def timestamps_add(url, timestamp, delta, checksum):
    '''Save new timestamps.  Do nothing if unchanged.
    '''
    ta = transactions.setdefault(url, [])
    if not len(ta) or timestamp > ta[-1][2]: # or delta != ta[-1][1]:
        ta.append([int(time.time()), delta, timestamp, checksum])
        timestamp_file.write('\t'.join([str(i) for i in ta[-1]+[url]])+'\n')
        timestamp_file.flush()

def timestamps_save_mirrors(repo_name, base_url, mirrors):
    '''Save new mirror -> repo data.  Do nothing if unchanged.
    '''
    mirrors_dict = {}.fromkeys(mirrors, base_url)
    if( base2repo.get(base_url, [None])[0] != repo_name
     or base2repo[base_url][1] != mirrors_dict ):
        base2repo[base_url] = [repo_name, mirrors_dict]
        mirror2base.update(mirrors_dict)
        timestamp_file.write('\t'.join(['*', repo_name, base_url]+mirrors)+'\n')
        timestamp_file.flush()

def do_time_stats(repos):
    '''Report median time behind and average estimated to update for each mirror
    (in repos) and note any frozen mirrors.
    '''
    report = []
    for mirror_url, trans_list in transactions.items():
        if not mirror2base.has_key(mirror_url):
            continue
        base_url = mirror2base[mirror_url]
        repo = base2repo[base_url][0]
        if not repo_match(repo, repos):
            continue
        deltas = [i[1] for i in trans_list]
        deltas.sort()
        median = deltas[len(deltas)/2]
        report.append([repo, mirror_url, median, (trans_list[-1][1] > 2*60*60*24)])
    report.sort()
    repo_name = None
    for i in report:
        if i[0] != repo_name:
            repo_name = i[0]
            print repo_name
        print '%1s%6.1f  %s' % (i[3] and "!" or " ", float(i[2])/(60*60*24), i[1])

def vprint(lvl, s):
    '''Print error messages controlled by verbosity.
    '''
    if lvl <= verbosity:
        sys.stderr.write(s)
        sys.stderr.write('\n')

def replace_vars(s):
    '''Replace $releasever and $basearch with values, returns string.
    '''
    s = s.replace('$releasever', releasever)
    s = s.replace('$basearch', basearch)        # .repo files
    s = s.replace('$ARCH', basearch)            # mirror lists
    s = s.replace('$us', us)                    # timestamp file name
    return s

def urljoin(*args):
    url = ''
    for arg in args:
        if url == '':
            url = arg
        else:
            if not url.endswith('/') and not arg.startswith('/'):
                url += '/'
            url += arg
    return url

def urlopen_timeout(url, timeout=timeout):
    '''Open url with timeout, per Josef Meile in comp.lang.python.  Returns
    (flo|None, None|err).  Timeouts don't seem to actually work.
    '''
    orig_timeout = socket.getdefaulttimeout()
    socket.setdefaulttimeout(timeout)
    ur = None
    err = None
    try:
        try:
            ur = urllib2.urlopen(url)
        except socket.timeout:
            # I don't see socket.timeout errors, so I don't know if they're working.
            err = 'timeout'
        except IOError, inst:
            err = 'IOError'
            normal = False
            if hasattr(inst, 'errno') and inst.errno:
                err += '(%s)' % inst.errno
                normal = True
            if hasattr(inst, 'strerror') and inst.strerror:
                err += ': "%s"' % inst.strerror
                normal = True
            if not normal:
                err += '? %s' % str(inst)
        except (urllib2.HTTPError, urllib2.URLError, socket.error):
            err = 'error "%s"' % sys.exc_info()[0]
    finally:
        socket.setdefaulttimeout(orig_timeout)
    return ur, err

def repo_match(repo, repos):
    '''Does repo match a glob in repos (True if no globs)?
    '''
    had_match = (len(repos) == 0)
    for rg in repos:
        if fnmatch(repo, rg):
            had_match = True;
            break;
    return had_match

def get_repodata(url_list, xml):
    '''Get (and possibly cache) repo data files, returns [(url, text|None), ...].
    '''
    class GetRepodataThread(threading.Thread):
        '''Get repo data from URL.  Passed list of mirrordata [url, full_url, text, msg,
        elapsed_time, err].  Sets text, msg, elapsed_time, err.
        '''
        def __init__(self, md):
            threading.Thread.__init__(self)
            self.mirrordata = md
        
        def run(self):
            full_url = self.mirrordata[1]
            text = None
            msg = 'Can\'t open'
            start_time = time.time()
            ur, err = urlopen_timeout(full_url)
            elapsed = time.time() - start_time
            if ur is not None:
                try:
                    if elapsed >= timeout:
                        msg = 'Too slow (%ds) opening' % elapsed
                    else:
                        msg = '?'
                        try:
                            text = ur.read()
                        except IOError:
                            msg = 'Can\'t read'
                        except socket.timeout:
                            elapsed = time.time() - start_time
                            msg = 'Timeout (%ds) reading' % elapsed
                finally:
                    ur.close()
            self.mirrordata[2] = text
            self.mirrordata[3] = msg
            self.mirrordata[4] = time.time() - start_time
            self.mirrordata[5] = err
        
    mirrordata_list = []
    threads = []
    for url in url_list:
        text = None
        msg = None
        full_url = urljoin(url, 'repodata', xml)
        vprint(3, '  %s' % full_url)
        if cache_map is not None:
            name = cache_map.get(full_url, None)
            if name == '':
                msg = 'Can\'t open'
            elif name:
                try:
                    f = open(name)
                    try:
                        text = f.read()
                    finally:
                        f.close()
                    vprint(4, '    from cache file "%s"' % name)
                    msg = '?'
                except:
                    pass
        md = [url, full_url, text, msg, 0, None]
        mirrordata_list.append(md)
        if not msg:
            thd = GetRepodataThread(md)
            thd.start()
            threads.append(thd)
    
    for thd in threads:
        thd.join()
    
    result_list = []
    for md in mirrordata_list:
        url, full_url, text, msg, elapsed_time, err = tuple(md)
        base_url = urlparse.urlparse(url)[1]
        
        if text is None:
            vprint(2, '  %s "%s"%s' % (msg, base_url, err and ':\n    '+err or ''))
            if cache_map is not None:
                vprint(4, '    to failed cache')
                cache_map[full_url] = ''
                cache_map_file.write('%s\t%s\n' % (full_url, ''))
                cache_map_file.flush()
        else:
            if elapsed_time > 10:
                vprint(2, '  Slow server (%d secs) "%s"' % (elapsed_time, base_url))
            if cache_map is not None:
                hdl, name = tempfile.mkstemp('', CACHEDATAPREFIX, cache_dir)
                vprint(4, '    to cache file "%s"' % name)
                cache_map[full_url] = name
                os.write(hdl, text)
                os.close(hdl)
                cache_map_file.write('%s\t%s\n' % (full_url, name))
                cache_map_file.flush()
        result_list.append((url, text))
    return result_list

def get_node_text(nodelist):
    '''Returns text of text nodes.'''
    tl = []
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE:
            tl.append(node.data)
    return ''.join(tl)

def extract_repomd(url, rd):
    '''Returns dict of dicts containing repomd metadata or None.'''
    if not rd:
        return None
    dd = {}
    dom = None
    try:
        try:
            dom = parseString(rd)
            for data in dom.getElementsByTagName('data'):
                ed = {}
                dd[data.attributes.get('type').value] = ed
                for name in ['timestamp', 'checksum', 'location']:
                    ed[name] = get_node_text(data.getElementsByTagName(name)[0].childNodes)
        except:
            vprint(2, '  Can\'t parse data from "%s"' % url)
            vprint(5, rd)
            return None
    finally:    
        if dom:
            dom.unlink()
    for data in ['primary', 'filelists', 'other']:
        if not dd.has_key(data):
            vprint(2, '  Missing %s data from "%s"' % (data, url))
            vprint(5, rd)
            return None
    return dd

def do_repo(repo, repos, bases, bases_disabled, mirrors, mirrors_disabled):
    if not repo_match(repo, repos):
        return
    
    if len(bases) == 0:
        bases = bases_disabled
    if len(mirrors) == 0 and len(bases) <= 1:
        mirrors = mirrors_disabled
    for m in mirrors:
        m = replace_vars(m)
        lines = None
        if cache_map is not None:
            name = cache_map.get(m, None)
            if name:
                try:
                    f = open(name)
                    try:
                        lines = f.readlines()
                    finally:
                        f.close()
                except:
                    pass
        ur = None
        if lines is None:
            ur, err = urlopen_timeout(m)
            if ur is None:
                vprint(1, 'Can\'t open mirrorlist "%s" for repo "%s":\n    %s' % (m, repo, err))
            try:
                lines = ur.readlines()
            except (IOError, socket.timeout):
                vprint(1, '  Can\'t read mirrorlist "%s" for repo "%s"' % (m, repo))
            else:
                if cache_map is not None:
                    hdl, name = tempfile.mkstemp('', CACHEMIRRORPREFIX, cache_dir)
                    cache_map[m] = name
                    os.write(hdl, ''.join(lines))
                    os.close(hdl)
                    cache_map_file.write('%s\t%s\n' % (m, name))
                    cache_map_file.flush()
        for l in lines:
            bases.append(l[:-1])
        if ur:
            ur.close()
    
    bases = [replace_vars(s) for s in bases]
    base = bases[0]
    del bases[0]
    
    vprint(1, 'Repo: %s' % repo)
    vprint(1, '  base:    %s' % base)
    vprint(1, '  mirrors: %d' % len(bases))
    vprint(5, '  bases:   %s %s' % (bases, bases_disabled))
    vprint(5, '  mirrors: %s %s' % (mirrors, mirrors_disabled))
    
    timestamps_save_mirrors(repo, base, bases)
    b, rd = get_repodata([base], 'repomd.xml')[0]
    base_data = extract_repomd(b, rd)
    if not base_data:
        return
    pd = base_data['primary']
    pts = int(pd['timestamp'])
    tsd = datetime.datetime.fromtimestamp(float(pts))
    vprint(2, '  ++          %10s = %d/%02d/%02d %02d:%02d:%02d' % (
         pts, tsd.year, tsd.month, tsd.day, tsd.hour, tsd.minute, tsd.second ))
    vprint(2, '        %s' % pd['checksum'])
    timestamps_add(base, pts, 0, pd['checksum'])
    
    good = 0
    mismatches = 0
    fails = 0
    rematchlist = [[base_data, 0, 0]]
    need_mirrors = []
    for b in bases:
        if transactions.has_key(b) and pts == transactions[b][-1][2]:
            good += 1
            rematchlist[0][1] += 1
            #print 'trans', b
        else:
            need_mirrors.append(b)
    mirror_data = get_repodata(need_mirrors, 'repomd.xml')
    for b, rd in mirror_data:
        mirror_data = extract_repomd(b, rd)
        if not mirror_data:
            fails += 1
            continue
        md = mirror_data['primary']
        mts = int(md['timestamp'])
        mtsd = datetime.datetime.fromtimestamp(float(mts))
        td = tsd - mtsd
        timestamps_add(b, mts, td.days*(60*60*24)+td.seconds, md['checksum'])
        days = td.days + float(td.seconds) / (60*60*24)
        had_rematch = False
        for rm in rematchlist:
            if rm[0] == mirror_data:
                rm[1] += 1
                had_rematch = True
        if not had_rematch:
            rematchlist.append([mirror_data, 1, days])
        if mirror_data == base_data:
            good += 1
        else:
            mismatches += 1
            vprint(2, '  <>%s' % b)
            vprint(2, '   %4s days  %10s = %d/%02d/%02d %02d:%02d:%02d' % (
                round(days, 1),
                mts, mtsd.year, mtsd.month, mtsd.day,
                mtsd.hour, mtsd.minute, mtsd.second ))
            vprint(2, '        %s' % md['checksum'])
            if pd == md:
                vprint(2, '      primary matches.  %s mismatch.  %s missing.' % (
                    [ k for k, v in mirror_data.items()
                        if base_data.has_key(k) and v == base_data[k] ],
                    [ k for k in base_data.keys() if not mirror_data.has_key(k) ] ))
    rematchlist.sort(key=lambda i: i[2])
    vprint(1, '  up-to-date: %3d    behind: %3d    failed: %3d' % (good, mismatches, fails))
    vprint(1, '  (#: age(days)): %s' % (['%d: %3.1f' % (i[1], i[2]) for i in rematchlist]))

def do_repofile(fn, repos):
    vprint(1, 'Repo File: "%s"' % fn)
    repo_re = re.compile('\s*(#?)\s*\[([\w-]+)\]')
    url_re = re.compile( '\s*(#?)\s*(\w+)\s*=\s*(\S*)')
    expecting_re = re.compile('\s+(#?)\s*(\S+)')
    repo = None
    expecting = None
    bases = []
    bases_disabled = []
    mirrors = []
    mirrors_disabled = []
    f = open(fn)
    for l in f:
        if expecting:
            expecting_m = expecting_re.match(l)
            if not expecting_m:
                expecting = None
            else:
                if expecting_m.group(1) != '#':
                    if expecting == 'base':
                        bases.append(expecting_m.group(2))
                    elif expecting == 'mirror':
                        mirrors.append(expecting_m.group(2))
                continue
        repo_m = repo_re.match(l)
        if repo_m:
            if repo:
                do_repo(repo, repos, bases, bases_disabled, mirrors, mirrors_disabled)
            repo = None
            if not repo_m.group(1):
                repo = repo_m.group(2)
                bases = []
                bases_disabled = []
                mirrors = []
                mirrors_disabled = []
            continue
        url_m = url_re.match(l)
        if url_m:
            if url_m.group(2) == 'baseurl':
                if not url_m.group(3):
                    expecting = 'base'
                elif url_m.group(1) == '#':
                    bases_disabled.append(url_m.group(3))
                else:
                    bases.append(url_m.group(3))
            if url_m.group(2) == 'mirrorlist':
                if not url_m.group(3):
                    expecting = 'mirror'
                elif url_m.group(1) == '#':
                    mirrors_disabled.append(url_m.group(3))
                else:
                    mirrors.append(url_m.group(3))
    if repo[0]:
        do_repo(repo, repos, bases, bases_disabled, mirrors, mirrors_disabled)
    f.close()

def cli():
    import getopt
    def get_usage():
        return _(
            """%s %s  check synchronization of yum repositories\n"""
            """\n"""
            """Usage:\n"""
            """  %s [-h] [-f <repofile>]... [-r <repo>]... [-n <version>] [-b <basearch>]\n"""
            """\n"""
            """Help Options:\n"""
            """  -h, --help        this message\n"""
            """  -f, --file        repo file or dir glob [%s]\n"""
            """  -r, --repo        repo glob [all repos]\n"""
            """  -n, --releasever  (fedora) release version [%s]\n"""
            """  -b, --basearch    (fedora) base architecture [%s]\n"""
            """  -t, --times       report update time stats instead of querying repos\n"""
            """  -v, --verbose     be more verbose [%d]\n"""
            """  -q, --quiet       be less verbose\n"""
            """\n"""
            """With no parameters checks all repos in %s.\n"""
            """Remember to quote any globbing.\n"""
            """Run "pydoc %s" for documentation.\n""" ) % (
            us, __version__, us, REPOFILE, RELEASEVER, BASEARCH, VERBOSITY, REPOFILE, us )
    global releasever, basearch, verbosity
    
    files = []
    repos = []
    caching = False
    time_stats = False
    
    try:
        opts, args = getopt.getopt( sys.argv[1:], 'hf:r:n:b:vqt',
            ['help', 'file=', 'repo=', 'releasever=', 'basearch=', 'verbose', 'quiet', 'cache', 'times'] )
        for opt, val in opts:
            if opt in ['--help', '-h']:
                print get_usage()
                sys.exit(0)
            elif opt in ['--file', '-f']:
                files.append(val)
            elif opt in ['--repo', '-r']:
                repos.append(val)
            elif opt in ['--releasever', '-n']:
                releasever = val
            elif opt in ['--basearch', '-b']:
                basearch = val
            elif opt in ['--times', '-t']:
                time_stats = True
            elif opt in ['--verbose', '-v']:
                verbosity += 1
            elif opt in ['--quiet', '-q']:
                verbosity -= 1
            elif opt in ['--cache']:
                caching = True
        
        if len(args):
            print get_usage()
            sys.exit(1)
        
    except getopt.error, (msg, opt):
        print '%s: %s' % (us, msg)
        print _("""Run '%s --help' to see a full list of available command line options.""" % us)
        sys.exit(1)
    
    if not len(files):
        files.append(REPOFILE)
        
    files2 = []
    for g in files:
        gl = glob(g)
        if not len(gl):
            print _("""No match for "%s" """ % g)
            continue
        for f in glob(g):
            if os.path.isfile(f):
                files2.append(f)
            elif os.path.isdir(f):
                for wroot, wdirs, wfiles in os.walk(f, topdown=True):
                    del wdirs[:]        # walk only one level!
                    for f2 in wfiles:
                        if f2.endswith('.repo'):
                            files2.append(os.path.join(wroot, f2))
            else:
                print _("""Not a file or dir: "%s" """ % f)
    
    if caching:
        global cache_map, cache_dir, cache_map_file
        temp_dir_list = glob(os.path.join(tempfile.gettempdir(), us + '.*'))
        if len(temp_dir_list):
            cache_dir = temp_dir_list[0]
        else:
            cache_dir = tempfile.mkdtemp('', us+'.')
        vprint(4, 'Cache dir: "%s"' % cache_dir)
        cache_map = {}
        cache_map_list = glob(os.path.join(cache_dir, CACHEMAPPREFIX+'*'))
        if len(cache_map_list):
            name = cache_map_list[0]
            cache_map_file = open(name, 'r+')
            for l in cache_map_file:
                k, v = l[:-1].split('\t')
                cache_map[k] = v
                vprint(5, '  %s: %s' % (k, v))
        else:
            hdl, name = tempfile.mkstemp('', CACHEMAPPREFIX, cache_dir)
            cache_map_file = os.fdopen(hdl, 'w')
    
    return files2, repos, time_stats

def main():
    files, repos, time_stats = cli()
    timestamps_load(os.path.expanduser(replace_vars(TIMESTAMPFILE)))
    
    if time_stats:
        do_time_stats(repos)
    else:
        for f in files:
            do_repofile( f, repos )
    
    if cache_map_file:
        cache_map_file.close()
    if timestamp_file:
        timestamp_file.close()

if __name__ == '__main__':
    main()
